/*========================================================================
     "catime" - time-dependent chemical-activation problem solver
                 for multiple-well system
      This is a part of "ssumes" -
        Steady-State Unimoluclar Master Equation Solver
             Copyright (c) 2002-2018 by A. Miyoshi, Hiroshima Univ.
                                  created: May   18, 2011
                                   edited: May   31, 2011
        - added arbitrary excitation
                                   edited: June  14, 2018
        - added output of rates & gsum
                                   edited: Sept.  3, 2018
                              last edited: Sept.  6, 2018
--------------------------------------------------------------------------
 usage: catime basename(.ext)
--------------------------------------------------------------------------
 - If no extension is given, default extension '.inp' is assumed
 - output is written in [basename]_catime_vec.csv
========================================================================*/

#include "ssulibc.h"

void outputHeader(ostream &os, listWells &lw);
void outputResults(ostream &os, listWells &lw, int ci, double T, double p,
 timeList &tl, listRateAW &lr);

/*===== main function ==================================================*/

int main(int argc, char *argv[]) {
  int inWell, ntkns, idxT, idxP, err3set = false, itkn, alphaTset = false;
  int condInit = false, outVec = false;
  double err3inp, a1000inp, aTEinp;
  string ctrlfln, basfln, extfln, csvfln, vecfln, slin, key;
  ifstream inf;
  ofstream cof, vcf;
  vector<string> tkns;
  tempList tlst;
  pressList plst;
  timeList timl;
  listRateAW lrat;
  connex cnx;
  well molecw;
  listWells lstw;
  umolProb *ump = new umolProb;   // large object cannot be an autovariable

  cout << "catime - "; showRevision();
  if (argc != 2) { cout << "usage: catime basename(.ext)\n"; return 1; }
  ctrlfln = argv[1];
  basfln = getBasePart(ctrlfln); extfln = getExtPart(ctrlfln);
  if (extfln == "") { ctrlfln = basfln + ".inp"; }
  csvfln = basfln + "_catime_out.csv";
  vecfln = basfln + "_catime_vec.csv";
  ump->probtype = umolProb::TimEv;

  // --- read in control input
  inf.open(ctrlfln.c_str());
  if (!inf) { cout << "failed to open [" << ctrlfln << "].\n"; return 1; }
  inWell = false;
  while (getlParToksKey(inf, key, tkns, ntkns, slin) == NORM) {
    if (inWell) {
      if (key == "}") {
        if (lstw.add(molecw) == ERRE) {
          cout << "Duplicated well-index [" << molecw.index << "].\n";
          return 1;
        }
        inWell = false;
      } else if (key == "index") {
        if (ntkns < 2) { cout << "No value for key [index].\n"; return 1; }
        molecw.index = atoi(tkns[1].c_str());
      } else if (key == "filename") {
        if (ntkns < 2) { cout << "No value for key [filename].\n"; return 1; }
        molecw.fname = tkns[1];
      } else if (key == "offset") {
        if (ntkns < 2) { cout << "No value for key [offset].\n"; return 1; }
        molecw.offset = atoi(tkns[1].c_str());
      } else if (key == "connect") {
        if (ntkns < 3)
          { cout << "Two values needed for key [connect].\n"; return 1; }
        cnx.toWellIdx = atoi(tkns[1].c_str());
        cnx.chan = atoi(tkns[2].c_str()) - 1;
        molecw.connect.push_back(cnx);
      } else if (key == "recombChan") {
        if (ntkns < 2) { cout << "No value for key [recombChan].\n"; return 1; }
        molecw.caWell = true;
        molecw.actchan = atoi(tkns[1].c_str()) - 1;
      } else if (key == "excitGrain") {
        if (ntkns < 2) { cout << "No value for key [excitGrain].\n"; return 1; }
        molecw.exWell = true;
        molecw.exgrain = atoi(tkns[1].c_str());
      } else if (key == "excitArbit") {
        if (ntkns < 2) { cout << "No value for key [excitArbit].\n"; return 1; }
        molecw.axWell = true;
        molecw.fnafx = tkns[1];
      } else if (key == "truncate") {
        if (ntkns < 2) { cout << "No value for key [truncate].\n"; return 1; }
        molecw.lowest = atoi(tkns[1].c_str());
      } else { cout << "Warning: Invalid key [" << key << "].\n"; }
    } else {
      if (key == "well{") {
        molecw.clear();
        molecw.index = lstw.autoIndex();
        inWell = true;
      } else if (key.substr(0, 4) == "temp") {      // temperature list input
        if (tlst.addFrom(slin) == ERRE) { return 1; }
      } else if (key.substr(0, 5) == "press") {     // pressure list input
        if (plst.addFrom(slin) == ERRE) { return 1; }
      } else if (key.substr(0, 4) == "time") {      // time list input
        if (timl.addFrom(slin) == ERRE) { return 1; }
      } else if (key == "topCut") {
        if (ntkns < 2) { cout << "No value for key [topCut].\n"; return 1; }
        lstw.topCut = atoi(tkns[1].c_str());
      } else if (key == "err3") {
        if (ntkns < 2) { cout << "No value for key [err3].\n"; return 1; }
        err3inp = atof(tkns[1].c_str()); err3set = true;
      } else if (key == "alphaTdep") {
        if (ntkns < 3) { cout << "No value for key [alphaTdep].\n"; return 1; }
        a1000inp = atof(tkns[1].c_str()); aTEinp = atof(tkns[2].c_str());
        alphaTset = true;
      } else if (key == "kEthInt") {
        if (ntkns < 2) { cout << "No value for key [kEthInt].\n"; return 1; }
        lstw.kEthInt = atof(tkns[1].c_str());
      } else if (key == "kEthOut") {
        if (ntkns < 2) { cout << "No value for key [kEthOut].\n"; return 1; }
        lstw.kEthOut = atof(tkns[1].c_str());
      } else if (key == "verbose") {
        if (ntkns < 2) { cout << "No value for key [verbose].\n"; return 1; }
        ump->verbose = atoi(tkns[1].c_str());
      } else if (key == "output") {
        if (ntkns < 2) { cout << "Warning: No value for key [output].\n"; }
        else {
          for (itkn = 1; itkn < ntkns; itkn++) {
            if (tkns[itkn] == "vector") {
              outVec = true;
            } else {
              cout << "Warning: Invalid value [" << tkns[1]
                   << "] for key [output].\n";
            }
          }
        }
      } else if (key == "condition") {
        if (ntkns < 2) { cout << "Warning: No value for key [condition].\n"; }
        else {
          if (tkns[1] == "const") {
            ump->timeevcond = umolProb::TEconst;
            condInit = false;
          } else if (tkns[1] == "init") {
            ump->timeevcond = umolProb::TEinit;
            condInit = true;
          } else {
            cout << "Warning: Invalid value [" << tkns[1]
                 << "] for key [condition].\n";
          }
        }
      } else if (key == "sizeLSDSRW") {
        if (ntkns < 2) { cout << "No value for key [sizeLSDSRW].\n"; return 1; }
        ump->sizeLSDSRW = atof(tkns[1].c_str());
      } else { cout << "Warning: Invalid key [" << key << "].\n"; }
    }
  }
  inf.close();

  // --- read in Unimol RRKMTH outputs
  if (lstw.readUnimol() == ERRE) { return 1; }
  if (lstw.checkCA() == ERRE) { return 1; }
  if (err3set) { lstw.setErr3(err3inp); }
  if (alphaTset) { lstw.setAlpPar(a1000inp, aTEinp); }
  if (tlst.empty()) { lstw.mergedTlist(tlst); }
  reverse(tlst.begin(), tlst.end());
  if (plst.empty()) { lstw.mergedPlist(plst); }
  reverse(plst.begin(), plst.end());
  if (timl.empty()) { timl.setDefault(); }
  timl.addZero();
  lstw.printParETP(cout, ump->verbose);
  lstw.printParCA(cout);

  // --- do calculations
  if (ump->initMW(lstw) == ERRE) { return 1; }
  cof.open(csvfln.c_str());
  if (!cof) { cout << "failed to open [" << csvfln << "].\n"; return 1; }
  if (outVec) {
    vcf.open(vecfln.c_str());
    if (!vcf) { cout << "failed to open [" << vecfln << "].\n"; return 1; }
  }
   // --- main loop
  outputHeader(cof, lstw);
  for (idxT = 0; idxT < tlst.size(); idxT++) {
    for (idxP = 0; idxP < plst.size(); idxP++) {
      cout << "T = " << tlst[idxT] << " [K], p = " << plst[idxP] << " [Torr]:\n";
      if (ump->setMatMW(lstw, tlst[idxT], plst[idxP]) == ERRE) { return 1; }
      if (ump->solveTimeEvMW(lstw, timl, lrat) == ERRE) { return 1; }
      outputResults(cof, lstw, condInit, ump->T, ump->p, timl, lrat);
      if (outVec) { ump->outputVecTimEvMW(vcf, lstw, timl); }
    }
  }
  if (outVec) { vcf.close(); }
  cof.close();
  return 0;
}

/*----- output header of a csv file ------------------------------------*/

void outputHeader(ostream &os, listWells &lw) {
  int ch, posw, toposw;
   // - 1st line
  os << ",,,,out-going-&-stab-fractions";
  for (ch = 0; ch < lw.nTotOCh + lw.size() - 1; ch++) { os << ','; }
  os << ",out-going-&-stab-rates";
  for (ch = 0; ch < lw.nTotOCh + lw.size(); ch++) { os << ','; }
  if (lw.nTotCh - lw.nTotOCh > 0) { os << ",internal-rates"; }
  for (ch = 0; ch < lw.nTotCh - lw.nTotOCh - 1; ch++) { os << ','; }
  os << ",internal-populations";
  for (posw = 1; posw < lw.size(); posw++) { os << ','; }
  os << endl;
   // - 2nd line
  os << ",,,";
  for (posw = 0; posw < lw.size(); posw++) {
    os << ",well" << lw[posw].index;
    for (ch = 0; ch < lw[posw].nOutCh; ch++) { os << ','; }
  }
  for (posw = 0; posw < lw.size(); posw++) {
    os << ",well" << lw[posw].index;
    for (ch = 0; ch < lw[posw].nOutCh; ch++) { os << ','; }
  }
  os << ",ktot";
  for (posw = 0; posw < lw.size(); posw++) {
    if (lw[posw].nchan - lw[posw].nOutCh > 0)
      { os << ",well" << lw[posw].index; }
    for (ch = 0; ch < lw[posw].nchan - lw[posw].nOutCh - 1; ch++)
      { os << ','; }
  }
  for (posw = 0; posw < lw.size(); posw++)
    { os << ",well" << lw[posw].index; }
  os << endl;
   // - 3rd line
  os << "T[K],P[Torr],t[s],gsum";
  for (posw = 0; posw < lw.size(); posw++) {
    for (ch = 0; ch < lw[posw].nchan; ch++) {
      if (lw[posw].chInf[ch] == -1) { os << ",f" << (ch + 1); }
      if ((lw[posw].caWell) && (ch == lw[posw].actchan)) { os << "(back)"; }
    }
    os << ",fstab";
  }
  for (posw = 0; posw < lw.size(); posw++) {
    for (ch = 0; ch < lw[posw].nchan; ch++) {
      if (lw[posw].chInf[ch] == -1) { os << ",k" << (ch + 1); }
      if ((lw[posw].caWell) && (ch == lw[posw].actchan)) { os << "(back)"; }
    }
    os << ",kstab";
  }
  os << ',';
  for (posw = 0; posw < lw.size(); posw++) {
    for (ch = 0; ch < lw[posw].nchan; ch++) {
      if ((toposw = lw[posw].chInf[ch]) >= 0)
        { os << ",k" << (ch + 1) << "(to" << lw[toposw].index << ')'; }
    }
  }
  for (posw = 0; posw < lw.size(); posw++) { os << ','; }
  os << endl;
}

/*----- output results to a csv file -----------------------------------*/

void outputResults(ostream &os, listWells &lw, int ci, double T, double p,
 timeList &tl, listRateAW &lr) {
  int itl, itl0 = 1, posw, ch;

  if (ci) { itl0 = 0; }
  for (itl = itl0; itl < tl.size(); itl++) {
    if (itl == itl0) { os << T << ',' << p; }
    else { os << ","; }
    os << ',' << tl[itl] << ',' << lr[itl].gnormTot;
    for (posw = 0; posw < lw.size(); posw++) {
      for (ch = 0; ch < lw[posw].nchan; ch++) {
        if (lw[posw].chInf[ch] == -1)
          { os << ',' << lr[itl][posw].k[ch] / lr[itl].koutTot; }
      }
      os << ',' << lr[itl][posw].kstab / lr[itl].koutTot;
    }
    for (posw = 0; posw < lw.size(); posw++) {
      for (ch = 0; ch < lw[posw].nchan; ch++) {
        if (lw[posw].chInf[ch] == -1)
          { os << ',' << lr[itl][posw].k[ch]; }
      }
      os << ',' << lr[itl][posw].kstab;
    }
    os << ',' << lr[itl].koutTot;
    for (posw = 0; posw < lw.size(); posw++) {
      for (ch = 0; ch < lw[posw].nchan; ch++) {
        if (lw[posw].chInf[ch] >= 0)
          { os << ',' << lr[itl][posw].k[ch]; }
      }
    }
    for (posw = 0; posw < lw.size(); posw++)
      { os << ',' << lr[itl][posw].fpop; }
    os << endl;
  }
}

